@@ -2,8 +2,8 @@
  * inode.c - Inode handling code. Originated from the Linux-NTFS project.
  *
  * Copyright (c) 2002-2005 Anton Altaparmakov
- * Copyright (c) 2002-2006 Szabolcs Szakacsits
- * Copyright (c) 2004-2005 Yura Pakhuchiy
+ * Copyright (c) 2002-2008 Szabolcs Szakacsits
+ * Copyright (c) 2004-2007 Yura Pakhuchiy
  * Copyright (c) 2004-2005 Richard Russon
  *
  * This program/include file is free software; you can redistribute it and/or
@@ -87,7 +87,7 @@ static ntfs_inode *__ntfs_inode_allocate(ntfs_volume *vol)
 {
 	ntfs_inode *ni;
 
-	ni = (ntfs_inode*)calloc(1, sizeof(ntfs_inode));
+	ni = (ntfs_inode*)ntfs_calloc(sizeof(ntfs_inode));
 	if (ni)
 		ni->vol = vol;
 	return ni;
@@ -152,23 +152,22 @@ static void __ntfs_inode_release(ntfs_inode *ni)
 ntfs_inode *ntfs_inode_open(ntfs_volume *vol, const MFT_REF mref)
 {
 	s64 l;
-	ntfs_inode *ni;
+	ntfs_inode *ni = NULL;
 	ntfs_attr_search_ctx *ctx;
-	int err = 0;
 	STANDARD_INFORMATION *std_info;
 
-	ntfs_log_trace("Entering for inode 0x%llx.\n", MREF(mref));
+	ntfs_log_enter("Entering for inode %lld\n", MREF(mref));
 	if (!vol) {
 		errno = EINVAL;
-		return NULL;
+		goto out;
 	}
 	ni = __ntfs_inode_allocate(vol);
 	if (!ni)
-		return NULL;
+		goto out;
 	if (ntfs_file_record_read(vol, mref, &ni->mrec, NULL))
 		goto err_out;
 	if (!(ni->mrec->flags & MFT_RECORD_IN_USE)) {
-		err = ENOENT;
+		errno = ENOENT;
 		goto err_out;
 	}
 	ni->mft_no = MREF(mref);
@@ -178,7 +177,6 @@ ntfs_inode *ntfs_inode_open(ntfs_volume *vol, const MFT_REF mref)
 	/* Receive some basic information about inode. */
 	if (ntfs_attr_lookup(AT_STANDARD_INFORMATION, AT_UNNAMED,
 				0, CASE_SENSITIVE, 0, NULL, 0, ctx)) {
-		err = errno;
 		ntfs_log_trace("Failed to receive STANDARD_INFORMATION "
 				"attribute.\n");
 		goto put_err_out;
@@ -203,7 +201,7 @@ ntfs_inode *ntfs_inode_open(ntfs_volume *vol, const MFT_REF mref)
 	if (!l)
 		goto put_err_out;
 	if (l > 0x40000) {
-		err = EIO;
+		errno = EIO;
 		goto put_err_out;
 	}
 	ni->attr_list_size = l;
@@ -214,7 +212,7 @@ ntfs_inode *ntfs_inode_open(ntfs_volume *vol, const MFT_REF mref)
 	if (!l)
 		goto put_err_out;
 	if (l != ni->attr_list_size) {
-		err = EIO;
+		errno = EIO;
 		goto put_err_out;
 	}
 get_size:
@@ -239,17 +237,16 @@ get_size:
 		}
 	}
 	ntfs_attr_put_search_ctx(ctx);
+out:	
+	ntfs_log_leave("\n");
 	return ni;
+
 put_err_out:
-	if (!err)
-		err = errno;
 	ntfs_attr_put_search_ctx(ctx);
 err_out:
-	if (!err)
-		err = errno;
 	__ntfs_inode_release(ni);
-	errno = err;
-	return NULL;
+	ni = NULL;
+	goto out;
 }
 
 /**
@@ -265,6 +262,8 @@ err_out:
  * If it is an extent inode, we disconnect it from its base inode before we
  * destroy it.
  *
+ * It is OK to pass NULL to this function, it is just noop in this case.
+ *
  * Return 0 on success or -1 on error with errno set to the error code. On
  * error, @ni has not been freed. The user should attempt to handle the error
  * and call ntfs_inode_close() again. The following error codes are defined:
@@ -276,17 +275,19 @@ err_out:
  */
 int ntfs_inode_close(ntfs_inode *ni)
 {
+	int ret = -1;
+	
 	if (!ni)
 		return 0;
 
-	ntfs_log_trace("Entering for inode 0x%llx.\n", (long long)ni->mft_no);
+	ntfs_log_enter("Entering for inode %lld\n", (long long)ni->mft_no);
 
 	/* If we have dirty metadata, write it out. */
 	if (NInoDirty(ni) || NInoAttrListDirty(ni)) {
 		if (ntfs_inode_sync(ni)) {
 			if (errno != EIO)
 				errno = EBUSY;
-			return -1;
+			goto err;
 		}
 	}
 	/* Is this a base inode with mapped extent inodes? */
@@ -295,7 +296,7 @@ int ntfs_inode_close(ntfs_inode *ni)
 			if (ntfs_inode_close(ni->extent_nis[0])) {
 				if (errno != EIO)
 					errno = EBUSY;
-				return -1;
+				goto err;
 			}
 		}
 	} else if (ni->nr_extents == -1) {
@@ -349,7 +350,10 @@ int ntfs_inode_close(ntfs_inode *ni)
 	}
 	
 	__ntfs_inode_release(ni);
-	return 0;
+	ret = 0;
+err:
+	ntfs_log_leave("\n");
+	return ret;
 }
 
 /**
@@ -380,7 +384,7 @@ int ntfs_inode_close(ntfs_inode *ni)
 ntfs_inode *ntfs_extent_inode_open(ntfs_inode *base_ni, const MFT_REF mref)
 {
 	u64 mft_no = MREF_LE(mref);
-	ntfs_inode *ni;
+	ntfs_inode *ni = NULL;
 	ntfs_inode **extent_nis;
 	int i;
 
@@ -390,7 +394,7 @@ ntfs_inode *ntfs_extent_inode_open(ntfs_inode *base_ni, const MFT_REF mref)
 		return NULL;
 	}
 	
-	ntfs_log_trace("Opening extent inode 0x%llx (base mft record 0x%llx).\n",
+	ntfs_log_enter("Opening extent inode %lld (base mft record %lld).\n",
 			(unsigned long long)mft_no,
 			(unsigned long long)base_ni->mft_no);
 	
@@ -411,20 +415,17 @@ ntfs_inode *ntfs_extent_inode_open(ntfs_inode *base_ni, const MFT_REF mref)
 				ntfs_log_perror("Found stale extent mft "
 					"reference mft=%lld",
 					(long long)ni->mft_no);
-				return NULL;
+				goto out;
 			}
-			return ni;
+			goto out;
 		}
 	}
 	/* Wasn't there, we need to load the extent inode. */
 	ni = __ntfs_inode_allocate(base_ni->vol);
 	if (!ni)
-		return NULL;
-	if (ntfs_file_record_read(base_ni->vol, le64_to_cpu(mref), &ni->mrec,
-			NULL)) {
-		ntfs_log_perror("ntfs_file_record_read failed #2");
+		goto out;
+	if (ntfs_file_record_read(base_ni->vol, le64_to_cpu(mref), &ni->mrec, NULL))
 		goto err_out;
-	}
 	ni->mft_no = mft_no;
 	ni->nr_extents = -1;
 	ni->base_ni = base_ni;
@@ -443,10 +444,13 @@ ntfs_inode *ntfs_extent_inode_open(ntfs_inode *base_ni, const MFT_REF mref)
 		base_ni->extent_nis = extent_nis;
 	}
 	base_ni->extent_nis[base_ni->nr_extents++] = ni;
+out:
+	ntfs_log_leave("\n");
 	return ni;
 err_out:
 	__ntfs_inode_release(ni);
-	return NULL;
+	ni = NULL;
+	goto out;
 }
 
 /**
@@ -663,6 +667,7 @@ err_out:
  */
 int ntfs_inode_sync(ntfs_inode *ni)
 {
+	int ret = 0;
 	int err = 0;
 
 	if (!ni) {
@@ -671,7 +676,7 @@ int ntfs_inode_sync(ntfs_inode *ni)
 		return -1;
 	}
 
-	ntfs_log_trace("Entering for inode %lld\n", (long long)ni->mft_no);
+	ntfs_log_enter("Entering for inode %lld\n", (long long)ni->mft_no);
 
 	/* Update STANDARD_INFORMATION. */
 	if ((ni->mrec->flags & MFT_RECORD_IN_USE) && ni->nr_extents != -1 &&
@@ -782,10 +787,13 @@ sync_inode:
 		}
 	}
 
-	if (!err)
-		return 0;
-	errno = err;
-	return -1;
+	if (err) {
+		errno = err;
+		ret = -1;
+	}
+	
+	ntfs_log_leave("\n");
+	return ret;
 }
 
 /**
@@ -1043,6 +1051,7 @@ int ntfs_inode_free_space(ntfs_inode *ni, int size)
 		 * find next, because we don't need such.
 		 */
 		while (ctx->ntfs_ino->mft_no != ni->mft_no) {
+retry:			
 			if (ntfs_attr_lookup(AT_UNUSED, NULL, 0, CASE_SENSITIVE,
 						0, NULL, 0, ctx)) {
 				err = errno;
@@ -1054,6 +1063,10 @@ int ntfs_inode_free_space(ntfs_inode *ni, int size)
 			}
 		}
 
+		if (ntfs_inode_base(ctx->ntfs_ino)->mft_no == FILE_MFT && 
+		    ctx->attr->type == AT_DATA)
+			goto retry;
+
 		record_size = le32_to_cpu(ctx->attr->length);
 
 		/* Move away attribute. */
